import argparse
import io
from collections import Counter

from akarsu.akarsu import Akarsu


def main() -> None:
    parser = argparse.ArgumentParser(
        description="New Generation Profiler based on PEP 669"
    )
    parser.add_argument("-v", "--version", action="version", version="0.1.0")
    parser.add_argument("-f", "--file", type=str, help="Path to the file")
    args = parser.parse_args()

    if file := args.file:
        with io.open(file) as fp:
            source = fp.read()
        events = Akarsu(source, args.file).profile()
        counter: Counter = Counter()

        print(f"{'Count':>10}{'Event Type':^20}{'Filename(function)':<50}")
        for event, count in Counter(events).most_common():
            event_type, file_name, func_name = event
            counter[event_type] += 1
            print(f"{count:>10}{event_type:^20}{f'{file_name}({func_name})':<50}")

        print(f"\nTotal number of events: {counter.total()}")
        for event_type, count in counter.most_common():
            print(f"  {event_type} = {count}")


if __name__ == "__main__":
    main()



import sys
import types
from typing import Any, Final, cast

TOOL: Final[int] = 2
PY_CALLABLES: Final[tuple] = (types.FunctionType, types.MethodType)
MONITOR = sys.monitoring  # type:ignore

EVENTS = MONITOR.events
TRACKED_EVENTS: Final[tuple[tuple[int, str], ...]] = (
    (EVENTS.PY_START, "PY_START"),
    (EVENTS.PY_RESUME, "RESUME"),
    (EVENTS.PY_THROW, "THROW"),
    (EVENTS.PY_RETURN, "PY_RETURN"),
    (EVENTS.PY_YIELD, "YIELD"),
    (EVENTS.PY_UNWIND, "UNWIND"),
    (EVENTS.C_RAISE, "C_RAISE"),
    (EVENTS.C_RETURN, "C_RETURN"),
    (EVENTS.EXCEPTION_HANDLED, "EXCEPTION_HANDLED"),
    (EVENTS.STOP_ITERATION, "STOP ITERATION"),
)
EVENT_SET: Final[int] = EVENTS.CALL + sum(ev for ev, _ in TRACKED_EVENTS)


class Akarsu:
    def __init__(self, code: str, file_name: str) -> None:
        self.code = code
        self.file_name = file_name

    def profile(self) -> list[tuple[str, str, str]]:
        events = []

        if code := self.code.strip():
            indented_code = "\n".join(f"\t{line}" for line in code.splitlines())
            source = f"def ____wrapper____():\n{indented_code}\n____wrapper____()"
            code = compile(source, self.file_name, "exec")  # type:ignore

            for event, event_name in TRACKED_EVENTS:

                def record(
                    *args: tuple[types.CodeType, int], event_name: str = event_name
                ) -> None:
                    code = cast(types.CodeType, args[0])
                    events.append((event_name, code.co_filename, code.co_name))

                MONITOR.register_callback(TOOL, event, record)

            def record_call(
                code: types.CodeType, offset: int, obj: Any, arg: Any
            ) -> None:
                file_name = code.co_filename
                if isinstance(obj, PY_CALLABLES):
                    events.append(("PY_CALL", file_name, obj.__code__.co_name))
                else:
                    events.append(("C_CALL", file_name, str(obj)))

            MONITOR.use_tool_id(TOOL, "Akarsu")
            MONITOR.register_callback(TOOL, EVENTS.CALL, record_call)
            MONITOR.set_events(TOOL, EVENT_SET)
            try:
                exec(code)
            except:
                pass
            MONITOR.set_events(TOOL, 0)
            MONITOR.free_tool_id(TOOL)

            events = [e for e in events[2:-3] if "____wrapper____" not in e[2]]

        return events



import textwrap
import unittest

from akarsu.akarsu import Akarsu


class TestNine(unittest.TestCase):
    def check_events(self, events, expected_events):
        for event, expected_event in zip(events, expected_events):
            self.assertEqual(event, expected_event)

    def test_profile_print(self):
        code = "print('Hello, world!')"
        events = Akarsu(code, "<string>").profile()
        expected_events = [("C_CALL", "<string>", "<built-in function print>")]
        self.check_events(events, expected_events)

    def test_profile_isinstance(self):
        code = "isinstance(1, int)"
        events = Akarsu(code, "<string>").profile()
        expected_events = [("C_CALL", "<string>", "<built-in function isinstance>")]
        self.check_events(events, expected_events)

    def test_profile_generator(self):
        code = "list(i for i in range(5))"
        events = Akarsu(code, "<string>").profile()
        expected_events = [
            ("C_CALL", "<string>", "<class 'range'>"),
            ("PY_CALL", "<string>", "<genexpr>"),
            ("C_CALL", "<string>", "<class 'list'>"),
            ("PY_START", "<string>", "<genexpr>"),
            ("YIELD", "<string>", "<genexpr>"),
            ("RESUME", "<string>", "<genexpr>"),
            ("YIELD", "<string>", "<genexpr>"),
            ("RESUME", "<string>", "<genexpr>"),
            ("YIELD", "<string>", "<genexpr>"),
            ("RESUME", "<string>", "<genexpr>"),
            ("YIELD", "<string>", "<genexpr>"),
            ("RESUME", "<string>", "<genexpr>"),
            ("YIELD", "<string>", "<genexpr>"),
            ("RESUME", "<string>", "<genexpr>"),
            ("PY_RETURN", "<string>", "<genexpr>"),
        ]
        self.check_events(events, expected_events)

    def test_profile_empty_code(self):
        code = ""
        events = Akarsu(code, "<string>").profile()
        expected_events = []
        self.check_events(events, expected_events)

    def test_profile_nested_functions(self):
        source = textwrap.dedent("""
            def foo():
                print("Hello, world!")
            def bar():
                foo()
            bar()
            """)
        events = Akarsu(source, "<string>").profile()
        expected_events = [
            ("PY_CALL", "<string>", "bar"),
            ("PY_START", "<string>", "bar"),
            ("PY_CALL", "<string>", "foo"),
            ("PY_START", "<string>", "foo"),
            ("C_CALL", "<string>", "<built-in function print>"),
            ("C_RETURN", "<string>", "foo"),
            ("PY_RETURN", "<string>", "foo"),
            ("PY_RETURN", "<string>", "bar"),
        ]
        self.check_events(events, expected_events)

    def test_profile_class(self):
        source = textwrap.dedent("""
            class C:
                def foo(self):
                    x = 1
            c = C()
            c.foo()
            """)
        events = Akarsu(source, "<string>").profile()
        expected_events = [
            ("C_CALL", "<string>", "<built-in function __build_class__>"),
            ("PY_START", "<string>", "C"),
            ("PY_RETURN", "<string>", "C"),
            ("PY_CALL", "<string>", "foo"),
            ("PY_START", "<string>", "foo"),
            ("PY_RETURN", "<string>", "foo"),
        ]
        self.check_events(events, expected_events)

    def test_profile_class_method(self):
        source = textwrap.dedent("""
            class MyClass:
                @classmethod
                def foo(cls):
                    print("Hello, world!")
            my_class = MyClass()
            my_class.foo()
            """)
        events = Akarsu(source, "<string>").profile()
        expected_events = [
            ("C_CALL", "<string>", "<built-in function __build_class__>"),
            ("PY_START", "<string>", "MyClass"),
            ("C_CALL", "<string>", "<class 'classmethod'>"),
            ("C_RETURN", "<string>", "MyClass"),
            ("PY_RETURN", "<string>", "MyClass"),
            ("PY_CALL", "<string>", "foo"),
            ("PY_START", "<string>", "foo"),
            ("C_CALL", "<string>", "<built-in function print>"),
            ("C_RETURN", "<string>", "foo"),
            ("PY_RETURN", "<string>", "foo"),
        ]
        self.check_events(events, expected_events)
